;;; -*- Mode: Lisp; Package: ELP -*-
;;;
;;; Lambda-prolog top-level stuff.
;;;
;;; Sccs Id @(#)lp-top.lisp	1.43 10/6/89
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;


(in-package :ELP) (use-package :ergolisp)

(export '(query process-query elp)) ; top used to be exported: use elp.

;;; The user interface is defined by the top-level continuation.  The one
;;; here presents the bindings to the evars (but not the tevars) and the
;;; remaining constraint if nonempty, and then asks whether the user wants
;;; more solutions.  If so, return (signaling failure).  If not, throw to
;;; query-success.  The function returns t or nil in analogy to standard
;;; Prolog's yes and no.


(defvar *work-ahead* t
  "If T, eLP will work ahead after the first solution is presented.")

(defvar *present-eager* nil
  "If T, and *work-ahead*, eLP will present successive solutions as soon as
they are found.")

(defun ms-time (internal-time)
  (round (* 1000 internal-time) internal-time-units-per-second) )

(defun query (sterm &optional all-solutions)
  "Take the given STERM query, convert it to a Gform, and solve it.  If optional
second argument is not nil, show all solutions, rather than prompting for more."
  ;; (declare-ftype query (Sterm Boolean) Boolean)
  ;; Clear out old evar and uvar identifiers.
  (let ((cumulative-user-time 0)
	(cumulative-unparse-time 0))
    (ident:ident-init)
    (ctag-init)
    (dlet ( ((values goal tevars evars uvars)
	     (tlang-call convert-squery sterm)) )
      (if (null uvars)
	  (format *elp-trace-output* "Solving ...")
	  (format *elp-trace-output*
		  "Solving with new parameter~:[~;s~] ~{ ~A~}  ..."
		  (rest uvars) uvars))
      (let ((absolute-time (get-internal-run-time)))
	(unwind-protect
	    (process-query
	     goal
	     (tlang-value empty-constraint)
	     #'(lambda (solution-con)
		 (let ((user-time (- (get-internal-run-time) absolute-time)))
		   (incf cumulative-user-time user-time)
		   (setq absolute-time (get-internal-run-time))
		   ;; Timing bug: the following wait gets charged to unparsing.
		   (when *working-ahead*
		     (if *present-eager*
			 (format *elp-terminal-io* " ;~%")
			 (if (more-answers-p)
			     t		; just present solution (see below).
			     ;; otherwise: throw out of here.
			     (query-success))))
		   (tlang-call present-solution tevars evars uvars
			       solution-con
			       *elp-terminal-io*)
		   (finish-output *elp-terminal-io*)
		   (let ((unparse-time (- (get-internal-run-time) absolute-time)))
		     (incf cumulative-unparse-time unparse-time)
		     (when *timings-after-each-solution*
		       (format *elp-terminal-io* "~&%% Time(ms): ~D (elp) + ~D (unparsing)"
			       (ms-time user-time)
			       (ms-time unparse-time)))
		     (cond ((or all-solutions
				(and *work-ahead*
				     (setq *working-ahead* t))
				(more-answers-p))
			    (setq absolute-time (get-internal-run-time))
			    (values))	; Signal failure
			   (t  (setq absolute-time (get-internal-run-time))
			       (query-success)))))))
	  (incf cumulative-user-time (- (get-internal-run-time) absolute-time))
	  (when *timings-totals*
	    (format *elp-terminal-io* "~&%% Total Time(ms): ~D (elp) + ~D (unparsing)."
		    (ms-time cumulative-user-time)
		    (ms-time cumulative-unparse-time))))))))

(defun more-answers-p ()
  "Returns (with t), if more answers are requested, with NIL otherwise."
  (let ((response (read-line *elp-terminal-io*)))
    (and (> (length response) 0)
	 (char= (schar response 0) #\;))))

(defun user-top-input-p ()
  "T, if the user has typed something."
  (listen *elp-terminal-io*))

(defun process-query (goal initial-con initial-sc)
  "Process a query, given GOAL, INITIAL-CON, and INITIAL-SC."
  (declare-ftype process-query (Gform Constraint Scont) Boolean)
  (catch 'query-success			; Success catcher
    (catch-cut top-ctag
      (solve goal
	     initial-con
	     top-ctag initial-sc))  
    nil))
  
;;; The following SB dependency should be moved.

(defvar *initfile-taken-p* nil)

(defun elp (&key (initfile :default))
  "The super-top-level.  To be called from Lisp."
  ;; Fix for Lucid 4.0 and probably more general
  ;; fp Fri Nov 22 17:57:32 1991
  (setq *elp-terminal-io* *terminal-io*)
  (when (boundp 'tools::*ess-version*)
    (format *elp-terminal-io* "~&eLP Version ~A ~A~%"
	    tools::*ess-version* tools::*ess-version-date*))
  (catch 'bye
    (let* ((initfilename (if (or (eq initfile t) (eq initfile :default))
			     (merge-pathnames
			      (merge-pathnames (user-homedir-pathname)
					       "elp.init")
			      *default-pathname-defaults*)
			     initfile))
	   (initquery
	    (if (and initfilename (or (not *initfile-taken-p*)
				      (eq initfile t)))
		(let ((initfilepath (probe-file initfilename)))
		  (if initfilepath
		      (format nil "take ~S." (namestring initfilepath))
		      nil))
		nil)))
      (setq *initfile-taken-p* t)
      (top initquery))
    (loop
     (format *elp-terminal-io* "You are at the topmost top-level.  Use  ?- bye.  to exit to Lisp~%")
     (top))))

(defun top (&optional (initquery-string nil))
  "ELP top-level read-solve-print loop."
  ;; The special predicate 'bye does a throw to 'bye.
  ;; The special predicate 'pop does a throw to 'top.
  (catch 'pop
    (catch 'end-of-file
      (when initquery-string
	(catch 'lp-error
	  (query (with-input-from-string (instream initquery-string)
		   (tlang-call parse-sterm instream))
		 t)))
      (retry-catch :top
	(loop
	 (catch 'lp-error		; For parsing errors
	   (format *elp-terminal-io* "~&?- ")
	   (let ((*working-ahead* nil))
	     (format
	      ;; I can't get a fresh-line to happen here.
	      ;; *elp-standard-input*
	      *elp-terminal-io*
	      "~&~a"
	      (if (query (tlang-call parse-sterm *elp-standard-input*))
		  "yes"
		  (if *working-ahead* "no more solutions" "no"))))))))))


